extern crate chrono;
use chrono::{Datelike, Timelike, DateTime, Weekday, TimeZone};

mod util;
use util::Increment;

//use std::default::Default;

pub enum Frequency {
    Yearly,
    Weekly,
    Monthly,
    Daily,
    Hourly,
    Minutely,
    Secondly
}

pub struct ByDay {
    pub num: i32,
    pub weekday: Weekday
}

pub struct Recurrence<Tz: TimeZone> {
    pub frequency: Frequency,
    pub count: Option<u32>,
    pub interval: Option<i32>,

    pub bysecond: Option<Vec<u32>>,
    pub byminute: Option<Vec<u32>>,
    pub byhour: Option<Vec<u32>>,
    pub byday: Option<Vec<ByDay>>,
    pub byweekno: Option<Vec<u32>>,
    pub bymonth: Option<Vec<u32>>,
    pub bymonthday: Option<Vec<u32>>,

    pub until: Option<DateTime<Tz>>
}

impl<Tz: TimeZone> Recurrence<Tz> {
    pub fn iter (&self, start: DateTime<Tz>) -> DateTimeIterator<Tz> {
        DateTimeIterator::new(self, start)
    }
}

pub struct DateTimeIterator<'a, Tz: TimeZone + 'a> {
    recurrence: &'a Recurrence<Tz>,
    start: DateTime<Tz>,
    next_datetimes: Vec<DateTime<Tz>>,
    count: u32
}

impl<'a, Tz: TimeZone + 'a> DateTimeIterator<'a, Tz> {
    fn new (recurrence: &Recurrence<Tz>, start: DateTime<Tz>) -> DateTimeIterator<Tz> {
        DateTimeIterator {
            recurrence: &recurrence,
            start: start,
            next_datetimes: vec![],
            count: 0
        }
    }

    fn generate_next_datetimes (&mut self) {
        self.next_datetimes = self.generate_by_month(
            self.generate_by_monthday(
                self.generate_by_hour(
                    self.generate_by_minute(
                        self.generate_by_second(vec![self.start.clone()])
                    )
                )
            )
        );

        let interval = self.recurrence.interval.unwrap_or(1);
        self.start = match self.recurrence.frequency {
            Frequency::Yearly => self.start.add_years(interval),
            Frequency::Monthly => self.start.add_months(interval as u32),
            Frequency::Weekly => self.start.add_weeks(interval as u32),
            Frequency::Daily => self.start.add_days(interval as u32),
            Frequency::Hourly => self.start.add_hours(interval as u32),
            Frequency::Minutely => self.start.add_minutes(interval as u32),
            Frequency::Secondly => self.start.add_seconds(interval as u32)
        };
    }

    fn generate_by_second (&self, starting_datetimes: Vec<DateTime<Tz>>) -> Vec<DateTime<Tz>> {
        if let Some(ref bysecond) = self.recurrence.bysecond {
            let mut next_datetimes = vec![];
            for start in starting_datetimes {
                for second in bysecond {
                    if let Some(datetime) = start.with_second(*second) {
                        next_datetimes.push(datetime);
                    }
                }
            }
            next_datetimes
        } else {
            starting_datetimes
        }
    }

    fn generate_by_minute (&self, starting_datetimes: Vec<DateTime<Tz>>) -> Vec<DateTime<Tz>> {
        if let Some(ref byminute) = self.recurrence.byminute {
            let mut next_datetimes = vec![];
            for start in starting_datetimes {
                for minute in byminute {
                    if let Some(datetime) = start.with_minute(*minute) {
                        next_datetimes.push(datetime);
                    }
                }
            }
            next_datetimes
        } else {
            starting_datetimes
        }
    }

    fn generate_by_hour (&self, starting_datetimes: Vec<DateTime<Tz>>) -> Vec<DateTime<Tz>> {
        if let Some(ref byhour) = self.recurrence.byhour {
            let mut next_datetimes = vec![];
            for start in starting_datetimes {
                for hour in byhour {
                    if let Some(datetime) = start.with_hour(*hour) {
                        next_datetimes.push(datetime);
                    }
                }
            }
            next_datetimes
        } else {
            starting_datetimes
        }
    }

    fn generate_by_monthday (&self, starting_datetimes: Vec<DateTime<Tz>>) -> Vec<DateTime<Tz>> {
        if let Some(ref bymonthday) = self.recurrence.bymonthday {
            let mut next_datetimes = vec![];
            for start in starting_datetimes {
                for monthday in bymonthday {
                    if let Some(datetime) = start.with_day(*monthday) {
                        next_datetimes.push(datetime);
                    }
                }
            }
            next_datetimes
        } else {
            starting_datetimes
        }
    }

    fn generate_by_month (&self, starting_datetimes: Vec<DateTime<Tz>>) -> Vec<DateTime<Tz>> {
        if let Some(ref bymonth) = self.recurrence.bymonth {
            let mut next_datetimes = vec![];
            for start in starting_datetimes {
                for month in bymonth {
                    if let Some(datetime) = start.with_month(*month) {
                        next_datetimes.push(datetime);
                    }
                }
            }
            next_datetimes
        } else {
            starting_datetimes
        }
    }
}

impl<'a, Tz: TimeZone + 'a> Iterator for DateTimeIterator<'a, Tz> {
    type Item = DateTime<Tz>;

    fn next (&mut self) -> Option<DateTime<Tz>> {
        if self.next_datetimes.is_empty() {
            self.generate_next_datetimes();
        }

        self.count = self.count + 1;
        if let Some(count) = self.recurrence.count {
            if count < self.count {
                return None;
            }
        }

        self.next_datetimes.pop()
    }
}

#[cfg(test)]
mod tests {
    use {Recurrence, Frequency};
    use chrono::DateTime;
    use chrono::offset::Local;

    #[test]
    fn test_daily_recurrence () {
        let recurrence = Recurrence {
            frequency: Frequency::Daily,
            count: Some(10),
            interval: Some(1),

            byhour: Some(vec![4]),
            bysecond: Some(vec![0]),
            byminute: Some(vec![20]),
            byday: None,
            byweekno: None,
            bymonth: None,
            bymonthday: None,

            until: None
        };

        for datetime in recurrence.iter(Local::now()) {
            println!("{:?}", datetime);
        }
    }

    #[test]
    fn test_yearly_recurrence () {
        let recurrence = Recurrence {
            frequency: Frequency::Yearly,
            count: Some(10),
            interval: Some(1),

            byhour: None,
            bysecond: None,
            byminute: None,
            byday: None,
            byweekno: None,
            bymonth: Some(vec![2]),
            bymonthday: Some(vec![14]),

            until: None
        };

        for datetime in recurrence.iter(Local::now()) {
            println!("{:?}", datetime);
        }
    }
}
