use chrono::{Datelike, DateTime, TimeZone, Timelike, NaiveDate};

pub trait Increment {
    fn add_years (&self, interval: i32) -> Self;
    fn add_months (&self, interval: u32) -> Self;
    fn add_weeks (&self, interval: u32) -> Self;
    fn add_days (&self, interval: u32) -> Self;
    fn add_hours (&self, interval: u32) -> Self;
    fn add_minutes (&self, interval: u32) -> Self;
    fn add_seconds (&self, interval: u32) -> Self;
}

impl<Tz: TimeZone> Increment for DateTime<Tz> {
    fn add_years (&self, interval: i32) -> DateTime<Tz> {
        self.with_year(self.year() + interval).unwrap()
    }

    fn add_months (&self, interval: u32) -> DateTime<Tz> {
        let months = self.month() + interval;
        if months >= 12 {
            self.with_month(months % 12).unwrap().add_years((months / 12) as i32)
        } else {
            self.with_month(months).unwrap()
        }
    }

    fn add_weeks (&self, interval: u32) -> DateTime<Tz> {
        self.add_days(interval * 7)
    }

    fn add_days (&self, interval: u32) -> DateTime<Tz> {
        let last_day = last_day_of_month(self.year(), self.month());
        let days = self.day() + interval;
        if days > last_day {
            self.with_day(days % last_day).unwrap().add_months(days / last_day)
        } else {
            self.with_day(days).unwrap()
        }
    }

    fn add_hours (&self, interval: u32) -> DateTime<Tz> {
        let hours = self.hour() + interval;
        if hours >= 24 {
            self.with_hour(hours % 24).unwrap().add_days(hours / 24)
        } else {
            self.with_hour(hours).unwrap()
        }
    }

    fn add_minutes (&self, interval: u32) -> DateTime<Tz> {
        let minutes = self.minute() + interval;
        if minutes >= 60 {
            self.with_minute(minutes % 60).unwrap().add_hours(minutes / 60)
        } else {
            self.with_minute(minutes).unwrap()
        }
    }

    fn add_seconds (&self, interval: u32) -> DateTime<Tz> {
        let seconds = self.second() + interval;
        if seconds >= 60 {
            self.with_second(seconds % 60).unwrap().add_minutes(seconds / 60)
        } else {
            self.with_second(seconds).unwrap()
        }
    }
}

// from https://github.com/chronotope/chrono/issues/29#issuecomment-84492746
pub fn last_day_of_month (year: i32, month: u32) -> u32 {
    NaiveDate::from_ymd_opt(year, month + 1, 1).unwrap_or(NaiveDate::from_ymd(year + 1, 1, 1)).pred().day()
}
